/*
 * Device Driver for EG&G ACE-MCS 913
 *
 * for Linux kernel 2.0.x
 */

static char rcs_revision[] = "$Revision: 0.6.0 $";

#ifdef MODULE
#include <linux/module.h>
#else
#error "eggmcs913 has to be compiled as a module"
#endif

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/mman.h>

#include <linux/timer.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/string.h>

#include <asm/dma.h>  /* MAX_DMA.. */


#include <linux/wrapper.h> /* for mem_map_reserve() ... */

#include "eggmcs913.h"

static eggmcs913cfg_t boardcfg;

static eggmcs913_t* boardlist;
static nboards = 0;
/* An element corresponds to a unique device address as selected
   with a jumper on the board. addr 1 -> element 0 ...
   It will be initialized with pointers into 'boardlist' */
static eggmcs913_t* boards[EGGMCS913_BOARDS_MAX] = {
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};


/* These routines allow a faked operation of the driver, to test
  applications, when the hardware in not present.

  A timer is used to generate a signal after a time corresponding to 
  the dwelltime. The channel-count is incremented, fake data is
  generated and if the passlength has been reached, an interrupt is
  simulated by calling the normal interrupt handler. */

static void eggmcs913_interrupt(int irq, void* dontknow,
				struct pt_regs *ignored);
static void mcs_start_timer(int boardid);

static void eggmcs913_timer_interrupt(unsigned long boardid)
{
	eggmcs913_t *boardPtr = boards[boardid];
	unsigned long *b = (unsigned long*) boardPtr->membase;
	int i;

	del_timer(&boardPtr->mcs_timer_list);

	if (boardPtr->cur_channel == boardPtr->passlength) {
		boardPtr->cur_channel = 0;
		eggmcs913_interrupt(boardcfg.irq, (void*) boardid, NULL);
	} else {
		if (boardPtr->add==1) {
			b[boardPtr->cur_channel]+=
				boardPtr->cur_channel%(10*(boardid+1))+1;
		} else {
			b[boardPtr->cur_channel]=
				boardPtr->cur_channel%(10*(boardid+1)) +
				boardPtr->passcount;
		}
		boardPtr->cur_channel++;
		mcs_start_timer(boardid);
	}
}

static void mcs_start_timer(int boardid)
{

	eggmcs913_t *boardPtr = boards[boardid];
	if (boardPtr->active == 0) {
		boardPtr->cur_channel = 0;
		return;
	}
	init_timer(&boardPtr->mcs_timer_list);
	boardPtr->mcs_timer_list.expires = jiffies + 2*(boardPtr->dwelltime+1);
	boardPtr->mcs_timer_list.data = boardid;
	boardPtr->mcs_timer_list.function = eggmcs913_timer_interrupt;
	add_timer(&boardPtr->mcs_timer_list);
}

/* end of fake routines */

static int calc_timers(unsigned int dwt, int* timer1, int* timer2,
		       int* remain)
{
	unsigned int  tmax = 0x10000;
	int tmin = 2;
	int f, r;

	int rmin;

	int i, imin;

	unsigned int ticks;


	ticks = 2*dwt;

	if (dwt == 0) {
		*timer1 = 0;
		*timer2 = 0;
		*remain = 0;
		return 0;
	}
	if (dwt < EGGMCS913_MIN_DWELLTIME) {
		printk("dwelltime = %d < %d, not allowed \n",\
		       dwt, EGGMCS913_MIN_DWELLTIME);
		return -EINVAL;
	}
	if (dwt > 2147418112) {
		printk("dwelltime = %d > %d, not allowed \n", dwt,
		       EGGMCS913_MAX_DWELLTIME);
		return -EINVAL;
	}
	if (dwt <tmax) {
		*timer1 = dwt;
		*timer2 = 2;
		*remain = 0;
		return 0;
	} else {
		rmin = tmax;
		imin = tmin;
		for (i = ticks/(tmax-1)+1; i< tmax; i++) {
			r = ticks%i;
			if (r<=rmin) {
				f = ticks/i;
				rmin = r;
				imin = i;
			}
		}

		*timer2 = ticks/imin;
		*timer1 = imin;
		*remain = ticks%imin;
        
		return 0;
	}
}

static inline void eggmcs913_board_select(int board)
{
	MCS_PRINTK("eggmcs913_board_select: %d.\n", board);

	/* select board */
	outb_p(board, EGGMCS913_ADDRESS(boardcfg.baseaddr) );
}

static int eggmcs913_set_irqhandler(void irqhndl(int irq, void*,
 struct pt_regs *ignored)) 
{
	if (boardcfg.irqhandler != NULL) {
		free_irq(boardcfg.irq, NULL);
		boardcfg.irqhandler = NULL;
	}

	if(request_irq(boardcfg.irq, irqhndl, SA_INTERRUPT, "eggmcs913",
		       NULL)) {
		printk("eggmcs913: set_irqhandler: Cannot request IRQ(%d).\n",
		       boardcfg.irq);
		return -(EIO);
	}

	boardcfg.irqhandler = irqhndl;

	return 0;
}

/* Set the default values. Only in the struct, the board is
   programmed  when started. */
static void eggmcs913_set_default(eggmcs913_t *boardPtr)
{
	MCS_PRINTK("eggmcs913_default: %d.\n", board);

	boardPtr->dwelltime       = 250;
	boardPtr->passlength      = 4;
	boardPtr->passcountpreset = 1;
	boardPtr->extstart        = 0;
	boardPtr->extdwelltime    = 0;
	boardPtr->add             = 1;

	boardPtr->passcount       = 0;
	boardPtr->cmd             = 0;
}

/*
  set_dwelltime
 
  dwelltime in usec.
 
  Program the counters for the dwelltime; there are 2 cascaded
  2-byte * counters. Their mode is */
static int eggmcs913_set_dwelltime(eggmcs913_t *boardPtr, long dwelltime)
{
	int cnt0, cnt1;

	MCS_PRINTK("eggmcs913_set_dwelltime: %d us = 0x%04x.\n",
		   dwelltime, dwelltime); 

	cnt0 = boardPtr->cnt0;
	cnt1 = boardPtr->cnt1;


	MCS_PRINTK("eggmcs913_set_dwelltime: cnt0 = %d  cnt1 = %d\n",
		   cnt0, cnt1);

	/* for 0xffff the counter has to be set to 0 */
	if (cnt0 == 0xffff) cnt0 = 0;
	if (cnt1 == 0xffff) cnt1 = 0;

	/* this should never happen */
	if (cnt0 < 2  || cnt1 < 2) {
		printk( "eggmcs913_set_dwelltime: ERROR cnt <2 !!!!!\n" );
		return -1;
	}

	/* set counters */

	eggmcs913_board_select(boardPtr->boardid);
	/* counter 1 (the write operations to control-word addressing
	   counter0 with different modes make the out0 go high/low
	   which is connected to clk1. This is necessary to
	   make cnt1 load the presets). */

	/* 0x30 = 001100110 :  counter0, twobyte transfer, mode 0 */
	outb_p( 0x30, EGGMCS913_TIM8253_CW(boardcfg.baseaddr) );
	/* 0x74 = 011100110 :  counter1, twobyte transfer, mode 2 */
	outb_p( 0x74, EGGMCS913_TIM8253_CW(boardcfg.baseaddr) );

	MCS_PRINTK("eggmcs913_set_dwelltime: cnt1 low  = 0x%02x\n",
		   LOWBYTE(cnt1));
	MCS_PRINTK("eggmcs913_set_dwelltime: cnt1 high = 0x%02x\n",
		   HIGHBYTE(cnt1));
	outb_p(  LOWBYTE(cnt1), EGGMCS913_TIM8253_C1(boardcfg.baseaddr) );
	outb_p( HIGHBYTE(cnt1), EGGMCS913_TIM8253_C1(boardcfg.baseaddr) );

	/* 0x34 = 001100100 :  counter0, twobyte transfer, mode 2 */
	outb_p( 0x34, EGGMCS913_TIM8253_CW(boardcfg.baseaddr) );
	/* 0x30 = 001100000 :  counter0, twobyte transfer, mode 0 */
	outb_p( 0x30, EGGMCS913_TIM8253_CW(boardcfg.baseaddr) );


	/* counter 0 */
	/* 0x34 = 001100100 :  counter0, twobyte transfer, mode 2 */
	outb_p( 0x34, EGGMCS913_TIM8253_CW(boardcfg.baseaddr) );

	MCS_PRINTK("eggmcs913_set_dwelltime: cnt0 low  = 0x%02x\n",
		   LOWBYTE(cnt0));
	MCS_PRINTK("eggmcs913_set_dwelltime: cnt0 high = 0x%02x\n",
		   HIGHBYTE(cnt0));
	outb_p( LOWBYTE(cnt0), EGGMCS913_TIM8253_C0(boardcfg.baseaddr));
	outb_p(HIGHBYTE(cnt0), EGGMCS913_TIM8253_C0(boardcfg.baseaddr));

	return 0;
}

static void eggmcs913_set_passlength(eggmcs913_t* boardPtr, int passlength)
{
	eggmcs913_board_select(boardPtr->boardid);
	MCS_PRINTK("eggmcs913_set_passlength: %d.\n", passlength);
	/* 0xb6 = 101100110 :  counter2, twobyte transfer, mode 3 */
	outb_p(0xb6, EGGMCS913_TIM8253_CW(boardcfg.baseaddr));

	outb_p( LOWBYTE(passlength), EGGMCS913_TIM8253_C2(boardcfg.baseaddr));
	outb_p(HIGHBYTE(passlength), EGGMCS913_TIM8253_C2(boardcfg.baseaddr));
}

static void eggmcs913_clear(eggmcs913_t* boardPtr)
{
	int i;
	unsigned long *b = (unsigned long*) boardPtr->membase;

	MCS_PRINTK("eggmcs913_clear:.\n");
	eggmcs913_board_select(boardPtr->boardid);

	if (boardcfg.fake != 0) {
		for (i=0; i<boardcfg.nchannels; i++) {
			/* b[i] = i; */
			b[i] = 0;
		}
	} else {
		for (i=0; i<boardcfg.nchannels; i++) {
			b[i] = 0;
		}
	}

	boardPtr->passcount       = 0;
}

static void eggmcs913_makecmd(eggmcs913_t* boardPtr)
{
	int cmd;

	MCS_PRINTK("eggmcs913_makecmd:.\n");

	cmd = 0;

	if (boardPtr->add) cmd += EGGMCS913_ADDMODE;
	if (boardPtr->extdwelltime) {
		cmd += EGGMCS913_EXTDWELLTIME;
	}

	if (boardPtr->extstart) cmd += EGGMCS913_EXTSTART;

	MCS_PRINTK("eggmcs913_makecmd:cmd = 0x%02x.\n", cmd);
	boardPtr->cmd = cmd;
}

static void eggmcs913_restart(eggmcs913_t* boardPtr)
{
	MCS_PRINTK("eggmcs913_restart:.\n");

	eggmcs913_board_select(boardPtr->boardid);
	if (boardPtr->extdwelltime) {
		/* 0x70 = 011100000 :  counter1, twobyte transfer,
		   mode 0 => OUT1=LOW */
		outb_p( 0x70, EGGMCS913_TIM8253_CW(boardcfg.baseaddr) );
	}

	MCS_PRINTK("eggmcs913_restart:cmd = 0x%02x.\n", boardPtr->cmd);

	outb_p( boardPtr->cmd|EGGMCS913_UNKNOWN_BIT,
		EGGMCS913_CONTROL(boardcfg.baseaddr) );
    
	if (boardcfg.fake != 0) {
		if (boardPtr->active == 1) {
			mcs_start_timer(boardPtr->boardid);
		}
		return;
	}
}

static int eggmcs913_start(eggmcs913_t* boardPtr)
{
	int cmd, stat = 0;

	MCS_PRINTK("eggmcs913_start:.\n");

	if (boardPtr->active == 1) {
		printk("eggmcs913_start: already running.\n");
		return 0;
	}

	if (boardPtr->passcount == boardPtr->passcountpreset) {
		MCS_PRINTK("eggmcs913: passcount == passcountpreset\n");
		return 0;
	}

	cmd = 0;

	if (boardPtr->add) cmd += EGGMCS913_ADDMODE;
	if (boardPtr->extdwelltime) {
		cmd += EGGMCS913_EXTDWELLTIME;
		/* 0x70 = 011100000 :  counter1, twobyte transfer,
		   mode 0 => OUT1=LOW */
		outb_p( 0x70, EGGMCS913_TIM8253_CW(boardcfg.baseaddr) );
	} else {
		if ((stat = eggmcs913_set_dwelltime(boardPtr, boardPtr->dwelltime)) <0) {
			return stat;
		}
	}

	eggmcs913_set_passlength(boardPtr, boardPtr->passlength);

	if (boardPtr->extstart) cmd += EGGMCS913_EXTSTART;

	MCS_PRINTK("eggmcs913_start:cmd = 0x%02x.\n", cmd);

	eggmcs913_makecmd(boardPtr);

	outb_p( cmd                      ,
		EGGMCS913_CONTROL(boardcfg.baseaddr) );
	outb_p( cmd|EGGMCS913_UNKNOWN_BIT,
		EGGMCS913_CONTROL(boardcfg.baseaddr) );
	outb_p( EGGMCS913_UNKNOWN_CMD    ,
		EGGMCS913_CONTROL(boardcfg.baseaddr) );
	outb_p( cmd|EGGMCS913_UNKNOWN_BIT,
		EGGMCS913_CONTROL(boardcfg.baseaddr) );


	boardPtr->active = 1;

	if (boardcfg.fake != 0) {
		mcs_start_timer(boardPtr->boardid);
	}

	return 0;
}

static int eggmcs913_status(eggmcs913_t* boardPtr)
{
	int status;

	eggmcs913_board_select(boardPtr->boardid);
	status = inb_p(EGGMCS913_CONTROL(boardcfg.baseaddr) );
	MCS_PRINTK("eggmcs913_status: 0x%02x.\n", status);
	return status;
}

static void eggmcs913_stop(eggmcs913_t* boardPtr)
{
	MCS_PRINTK("eggmcs913_stop:.\n");
	eggmcs913_board_select(boardPtr->boardid);
	outb_p( EGGMCS913_UNKNOWN_CMD, EGGMCS913_CONTROL(boardcfg.baseaddr) );

	/* It should be checked whether the board is really
	   deactivated */
	boardPtr->active = 0;
}

static void eggmcs913_interrupt(int irq, void* dontknow,
				struct pt_regs *ignored)
{
	int status, bid, bids[EGGMCS913_BOARDS_MAX];
	eggmcs913_t *boardPtr;
	int i;

        /* in interrupt ... */

	bids[0] = -1;
        /* Find out which board caused the interrupt */
	if (boardcfg.fake != 0) {
		bids[0] = (int) dontknow;
		bids[1] = -1;
	} else {
		/* Check all (active) boards */
		i = 0;
		for (boardPtr = boardlist; boardPtr < boardlist + nboards;
		     boardPtr++) {
			if (boardPtr->active != 1) {
				continue;
			}
			/* select board */
			outb_p(boardPtr->boardid,
			       EGGMCS913_ADDRESS(boardcfg.baseaddr) );
			/* get status */
			status = inb_p(EGGMCS913_CONTROL(boardcfg.baseaddr) );
                        if (status & 0x80) {
				bids[i] = boardPtr->boardid;
			}
			i++;
		}
		bids[i] = -1;
	}

        
	if (bids[0] == -1) {
		printk("eggmcs913: interrupt - can't find the board, "
		       "that caused interrupt !!!\n");
		return;
	}

	for (i=0; i<EGGMCS913_BOARDS_MAX; i++) {
                if (bids[i] == -1) break;

                bid = bids[i];
		boardPtr = boards[bid];
		/*		printk("eggmcs913: boardid = %d\n", bid);*/

		boardPtr = boards[bid];
		boardPtr->passcount++;

		status = inb_p(EGGMCS913_CONTROL(boardcfg.baseaddr) );
		if (boardcfg.fake != 0) {
			status = 0;
		}

		MCS_PRINTK("eggmcs913: passcount = %ld.\n",
			   boardPtr->passcount);
		MCS_PRINTK("eggmcs913: passcountpreset = %ld.\n",
			   boardPtr->passcountpreset);

		if (status & EGGMCS913_EXTSTOP) {
			MCS_PRINTK("eggmcs913: interrupt, extstop requested\n");
			eggmcs913_stop(boardPtr);
			return;
		}
    
		if (boardPtr->stop_req) {
			MCS_PRINTK("eggmcs913: interrupt, stop requested\n");
			eggmcs913_stop(boardPtr);
			return;
		}
		if (boardPtr->passcount == boardPtr->passcountpreset) {
			MCS_PRINTK("eggmcs913: passcount == passcountpreset\n");
			eggmcs913_stop(boardPtr);
		} else {
			outb_p(EGGMCS913_UNKNOWN_CMD,
			       EGGMCS913_CONTROL(boardcfg.baseaddr) );
			eggmcs913_restart(boardPtr);
		}
	}
}

static int eggmcs913_read(struct inode *inode, struct file *filp,
                          char *buffer, int count)
{
	EGGMCS913_ClientData* mcs = (EGGMCS913_ClientData*) filp->private_data;
        eggmcs913_t* boardPtr = mcs->theBoard;
	int stat;
	/*
	  const unsigned long mask = 0xFFFFFFL;
	  */
	char* b = (char*) boardPtr->membase;
	char* bb;

	MCS_PRINTK("eggmcs913_read: reading out 0x%x bytes from 0x%lx.\n",
		   count, boardPtr->membase);


	bb = (char *) b;
	bb += (EGGMCS913_BYTESPERCHANNEL) * mcs->startchannel;
      
	if (count<0) return -(EINVAL);
         
	if (bb+count>b+boardcfg.memsize) return -(EINVAL);
	/*
	  if (bb+count>b+boardPtr->memsize) count = boardPtr->memsize;
	  */
         
	if ( (stat = verify_area(VERIFY_WRITE, (char *)buffer, count )) != 0) {
		return stat;
	}
	eggmcs913_board_select(boardPtr->boardid);
	memcpy_tofs(buffer, bb, count);
    
	return count;
}

static int eggmcs913_ioctl(struct inode *inode, struct file *filp,
                           unsigned int cmd, unsigned long arg)
{
	unsigned int stat;
	EGGMCS913_ClientData* mcs = (EGGMCS913_ClientData*) filp->private_data;
        eggmcs913_t* boardPtr = mcs->theBoard;

	switch (cmd)
		{
		case EGGMCS913_IOCTL_SET_DWELLTIME: {
			int cnt0, cnt1, remain;
			int dwelltime = (int) arg;

			if (mcs->controler != 1) {
				return -(EPERM);
			}

			if (dwelltime < EGGMCS913_MIN_DWELLTIME ||
			    dwelltime > EGGMCS913_MAX_DWELLTIME ) {
				return -(EINVAL);
			}
			boardPtr->dwelltime = dwelltime;
			if ( (stat = calc_timers(dwelltime, &cnt0, &cnt1,
						 &remain)) < 0) {
				printk( "eggmcs913_set_dwelltime: "
					"ERROR cnt <2 !!!!!\n" );
				return stat;
			}
			boardPtr->cnt0 = cnt0;
			boardPtr->cnt1 = cnt1;
		}
		break;


		case EGGMCS913_IOCTL_SET_PASSLENGTH: {
			int passlength = (int) arg;

			if (mcs->controler != 1) {
				return -(EPERM);
			}

			if (passlength < EGGMCS913_MIN_PASSLENGTH ||
			    passlength > EGGMCS913_MAX_PASSLENGTH ) {
				return -(EINVAL);
			}
			boardPtr->passlength = passlength;
		}
		break;

		case EGGMCS913_IOCTL_SET_PASSCOUNTPRESET: {
			unsigned long passcountpreset = (unsigned long) arg;

			if (mcs->controler != 1) {
				return -(EPERM);
			}

			if ( passcountpreset < 1 ) {
				return -(EINVAL);
			}
			boardPtr->passcountpreset = passcountpreset;
		}
		break;


		case EGGMCS913_IOCTL_START:

			if (mcs->controler != 1) {
				return -(EPERM);
			}

			stat = eggmcs913_start(boardPtr);
			return stat;
			break;
		case EGGMCS913_IOCTL_STOP:

			if (mcs->controler != 1) {
				return -(EPERM);
			}

			eggmcs913_stop(boardPtr);
			break;
		case EGGMCS913_IOCTL_STOPREQ:

			if (mcs->controler != 1) {
				return -(EPERM);
			}

			boardPtr->stop_req = (int)arg;
			break;

		case EGGMCS913_IOCTL_CLEAR:

			if (mcs->controler != 1) {
				return -(EPERM);
			}

			eggmcs913_clear(boardPtr);
			break;

		case EGGMCS913_IOCTL_SET_EXTDWELLTIME:

			if (mcs->controler != 1) {
				return -(EPERM);
			}

			boardPtr->extdwelltime = (int)arg;
			break;
		case EGGMCS913_IOCTL_SET_EXTSTART:

			if (mcs->controler != 1) {
				return -(EPERM);
			}

			boardPtr->extstart = arg;
			break;
		case EGGMCS913_IOCTL_SET_ADD:

			if (mcs->controler != 1) {
				return -(EPERM);
			}

			boardPtr->add = arg;
			break;



		case EGGMCS913_IOCTL_GET_DWELLTIME: {
			int dwelltime = boardPtr->cnt0 * boardPtr->cnt1 /2;
			if ( (stat = verify_area(VERIFY_WRITE, (char *)arg,
						 sizeof(long) )) != 0)
				return stat;

			put_user(dwelltime, (long *) arg);
		}
		break;
		case EGGMCS913_IOCTL_ACTIVE:
			if ( (stat = verify_area(VERIFY_WRITE, (char *)arg,
						 sizeof(long) )) != 0)
				return stat;

			put_user(boardPtr->active, (long *) arg);
			break;
		case EGGMCS913_IOCTL_GET_PASSCOUNT:
			if ((stat = verify_area(VERIFY_WRITE, (char *)arg,
						sizeof(unsigned long))) != 0){
				return stat;
			}

			put_user(boardPtr->passcount, (unsigned long *)arg);
			break;
		case EGGMCS913_IOCTL_GET_EXTDWELLTIME:
			if ( (stat = verify_area(VERIFY_WRITE, (char *)arg,
						 sizeof(long) )) != 0) {
				return stat;
			}

			put_user(boardPtr->extdwelltime, (long *) arg);
			break;
		case EGGMCS913_IOCTL_GET_EXTSTART:
			if ( (stat = verify_area(VERIFY_WRITE, (char *)arg,
						 sizeof(long) )) != 0)
				return stat;

			put_user(boardPtr->extstart, (long *) arg);
			break;
		case EGGMCS913_IOCTL_GET_ADD:
			if ( (stat = verify_area(VERIFY_WRITE, (char *)arg,
						 sizeof(long) )) != 0)
				return stat;

			put_user(boardPtr->add, (long *) arg);
			break;

		case EGGMCS913_IOCTL_NCHANNELS:
			if ( (stat = verify_area(VERIFY_WRITE, (char *)arg,
						 sizeof(long) )) != 0)
				return stat;

			put_user(boardcfg.nchannels, (long *) arg);
			break;

		case EGGMCS913_IOCTL_STARTCHANNEL:
			mcs->startchannel = (int)  arg;
			break;

		case EGGMCS913_IOCTL_STATUS:
			if ( (stat = verify_area(VERIFY_WRITE, (char *)arg,
						 sizeof(long) )) != 0)
				return stat;

			put_user(eggmcs913_status(boardPtr), (long *) arg);
			break;

		default: 
			MCS_PRINTK("eggmcs913: unknown IOCTL: %d\n.", cmd);
			return -(EINVAL);
			break;
		}

	return 0;
}

/* In non-fake mode there is no check that the correct board is
   selected, when a process accesses the mapped memory. This is ok
   if there are mmaps for one board. I do not know how to check
   for an existing mapping. The driver is not notified when a
   mapping goes away.

   Maybe do it via an ioctl-interface free/malloc -> LJ */
static int eggmcs913_mmap(struct inode *inode, struct file *filp,
			  struct vm_area_struct *vma)
{
	EGGMCS913_ClientData* mcs = (EGGMCS913_ClientData*) filp->private_data;
        eggmcs913_t* boardPtr = mcs->theBoard;

	printk("eggmcs913: offset = %lx\n", vma->vm_offset);

	if (remap_page_range(vma_get_start(vma),
			     virt_to_phys(boardPtr->membase),/*XXX fake-mode?*/
			     vma_get_end(vma) - vma_get_start(vma),
			     vma_get_page_prot(vma))) {
		printk("eggmcs913: remap bad\n");
		return -EAGAIN;
	}

	vma_set_inode(vma, inode);
	inode_inc_count(inode);

	printk("eggmcs913: remap ok\n");

	return 0;
}

static int eggmcs913_open(struct inode *inode, struct file *filp)
{
	int minor = MINOR(inode->i_rdev);
	int boardid;
	int mode;
	int controler;
	EGGMCS913_ClientData* mcs;
        eggmcs913_t* boardPtr;

        if (minor < 0 || minor > 7) {
		printk("eggmcs913: bad minor # %d\n", minor);
		return -EINVAL;
	}
	boardid = minor;

        if ( boards[boardid] == NULL) {
		return -ENODEV;
	}

	mode = filp->f_flags & O_ACCMODE;

	boardPtr = boards[boardid];
	printk("eggmcs913: open() boardid # %d\n", boardPtr->boardid);
	if (boardPtr->controler == 1) {
		if (mode==O_RDWR) {
			return -(EBUSY);
		}
		if (mode!=O_RDONLY) {
			return -(EINVAL);
		}
		controler = 0;
	} else {
		if (mode!=O_RDONLY) {
			controler = 1;
		} else {
			controler = 0;
		}
	}

	mcs = (EGGMCS913_ClientData*) kmalloc(sizeof(EGGMCS913_ClientData),
					      GFP_KERNEL);
	if (mcs == NULL) {
		return -(ENOMEM);
	}
	filp->private_data = mcs;

	mcs->controler    = controler;
	mcs->startchannel = 0;
	mcs->theBoard = boardPtr;

	if (mcs->controler == 1) {

		if (eggmcs913_set_irqhandler(eggmcs913_interrupt) != 0) {
			printk("eggmcs913: open: "
			       "Could not install irq-handler.\n");
			return -(EINVAL);
		}
		eggmcs913_set_default(boardPtr);
		/* eggmcs913_board_select(boardPtr->boardid); */
		boardPtr->controler = controler;
	}

	MOD_INC_USE_COUNT;

	boardPtr->usecount++;
	MCS_PRINTK("eggmcs913_open: EGGMCS913 opened.\n");

	return 0;
}

void eggmcs913_release(struct inode *inode, struct file *filp)
{
	EGGMCS913_ClientData* mcs = (EGGMCS913_ClientData*) filp->private_data;
        eggmcs913_t* boardPtr = mcs->theBoard;

	if (boardPtr->active==1 && boardPtr->usecount == 1) {
		eggmcs913_stop(boardPtr);
		if (boardcfg.fake != 0) {
			/* to get rid of a running timer */
			eggmcs913_timer_interrupt(boardPtr->boardid); 
			boardPtr->cur_channel = 0;
		}
	}

	if (mcs->controler==1) {
		boardPtr->controler = 0;
	}

	kfree(mcs);

	MOD_DEC_USE_COUNT;
	boardPtr->usecount--;

	MCS_PRINTK("eggmcs913_release: usecount = %d.\n", boardPtr->usecount);
}

struct file_operations eggmcs913_fops = {
	NULL,               /* seek */
	eggmcs913_read,     /* read */
	NULL,               /* write */
	NULL,               /* readdir */
	NULL,               /* select */
	eggmcs913_ioctl,    /* ioctl */
	eggmcs913_mmap,     /* mmap */
	eggmcs913_open,     /* open */
	eggmcs913_release   /* release */
};


#ifdef MODULE    /* only module version */

/* These can be set from the commandline, when loading the module. */
static int major = EGGMCS913_MAJOR;

/* irq, io and mem are for _all_ boards. If 2 boards have at least
one of these parameters in common, they have to be handled by the
same driver.

Assume that _all_ parameters are identical for _all_ boards
this drivers handles.
*/

/* A list of unique board-ids.
The MCS device addresses (1-8) all other uses of boardid are
0-based: 0-7 !!! */
static int boardids[EGGMCS913_BOARDS_MAX] ={ 0, 0, 0, 0, 0, 0, 0, 0};

static int irq = EGGMCS913_IRQ;
static int io = EGGMCS913_BASEADDR;
static int mem = EGGMCS913_MEMBASE;

static int fake = 0;


/* Name and version # of the driver, the name is used by the
 load-script to find a dynamically assigned major device #
 (one word) */
static char* eggmcs913_name = "eggmcs913-0.6.0";


int init_module(void)
{
	int stat, i;
	int err;
	eggmcs913_t *boardPtr;

        stat = 0;

	printk("eggmcs913: EG&G ACE-MCS 913 driver, version %s\n",
	       rcs_revision);

	boardcfg.fake            = 0;
	boardcfg.irq             = irq;
	boardcfg.baseaddr        = io;
	boardcfg.membase         = (unsigned long*) mem;
	
	boardcfg.fake            = fake;

	boardcfg.bytesperchannel = EGGMCS913_BYTESPERCHANNEL;
	boardcfg.nchannels       = EGGMCS913_CHANNELS;
	boardcfg.memsize         =
		boardcfg.bytesperchannel*boardcfg.nchannels;

	boardcfg.irqhandler      = NULL;


        nboards = 0;
	for (i=0; i<EGGMCS913_BOARDS_MAX; i++) {
                int bid = boardids[i];
		if (bid == 0) break;
		if (bid<1 || bid > EGGMCS913_BOARDS_MAX) {
			printk("eggmcs913: invalid device address: %d\n", bid);
			return -(EINVAL);
		}
		/* skip already existing ids */
                if (boards[bid-1] != NULL) {
			printk("eggmcs913: duplicate device address: %d\n",
			       bid);
			return -(EINVAL);
		}
                /* Just mark the entry */
                boards[bid-1] = (eggmcs913_t*) 1; 
                nboards++;
	}

	boardlist = (eggmcs913_t *)kmalloc( sizeof(eggmcs913_t) * nboards,
					    GFP_KERNEL);
	if (boardlist == NULL) {
		return -(ENOMEM);
	}

        boardPtr = boardlist;
	for (i=0; i<EGGMCS913_BOARDS_MAX; i++) {
                if (boards[i] == NULL) continue;

		boards[i] = boardPtr;

		boardPtr->boardid         = i;
		boardPtr->usecount        = 0;
		boardPtr->controler       = 0;

		boardPtr->stop_req        = 0;
		boardPtr->active          = 0;

		if (boardcfg.fake != 0) {
			boardPtr->membase = NULL;
		} else {
			boardPtr->membase = boardcfg.membase;
		}

		boardPtr->sz          = 0;

		boardPtr++;
	}


	if (boardcfg.fake != 0) {
		for (boardPtr = boardlist; boardPtr < boardlist + nboards;
		     boardPtr++) {
			char *start_addr, *end_addr;
			int sz, size;

			/* In fake mode every board has its own memory */
		

			/* mmapable */
			/* Taken from the sounddriver. It's not clear to
			   me what the requirements for these pages are
			   to make the mmap work, but this one does.
			   DMA-pages are not really necessary */

			for (sz = 0, size = PAGE_SIZE; size < boardcfg.memsize;
			     sz++, size <<= 1);

			printk("eggmcs913: sz = %d\n", sz);

			boardcfg.memsize = PAGE_SIZE * (1 << sz);
			start_addr =  __get_free_pages(GFP_KERNEL, sz,
						       MAX_DMA_ADDRESS);
			if (start_addr == NULL) {
				printk("eggmcs913: __get_free_pages failed\n");
				stat =  -(ENOMEM);
				goto fail_region;
			}
			end_addr = start_addr + 0x4000 - 1;
			for (i = MAP_NR(start_addr); i <= MAP_NR(end_addr); i++) {
				mem_map_reserve (i);
			}
			boardPtr->membase = start_addr;
			boardPtr->sz          = sz;
		} 
	}


	if (check_region(boardcfg.baseaddr, EGGMCS913_IOSIZE) != 0) {
		printk("eggmcs913: init_module: "
		       "I/O ports 0x%lx to 0x%lx already in use.\n",
		       boardcfg.baseaddr,
		       boardcfg.baseaddr + EGGMCS913_IOSIZE-1);
		goto fail_region;
	}

	request_region(boardcfg.baseaddr, EGGMCS913_IOSIZE, "eggmcs913");

	/* How to reserve the memory at  boardcfg.membase ? */


	err = register_chrdev(major, eggmcs913_name, &eggmcs913_fops);
	if (err < 0) {
		printk("eggmcs913: init_module: "
		       "Cannot register to major device %d.\n", major);
		goto fail_major;
	}
	if (major == 0) {
		major = err; /* dynamic major # */
	}

	for (boardPtr = boardlist; boardPtr < boardlist + nboards;
	     boardPtr++) {
		/* select board */
		outb_p(boardPtr->boardid, EGGMCS913_ADDRESS(boardcfg.baseaddr) );
		/* init, and reset it  ??? */
		outb_p(EGGMCS913_UNKNOWN_CMD,
		       EGGMCS913_CONTROL(boardcfg.baseaddr));
		eggmcs913_clear(boardPtr);
	}

	printk("eggmcs913: installed.\n");
        if (nboards == 0) {
		printk("eggmcs913: NO boards.\n");
	} else {
		printk("eggmcs913: %d boards.\n", nboards);
		printk("eggmcs913: "
		       "fake(%d), major(%d), irq(%d), baseaddr(0x%lx), "
		       "membase(0x%p).\n",
		       boardcfg.fake, major, boardcfg.irq,
		       boardcfg.baseaddr, boardcfg.membase);
		for (boardPtr = boardlist; boardPtr < boardlist + nboards;
		     boardPtr++) {
			printk("eggmcs913: (%d) membase = 0x%p\n",
			       boardPtr->boardid,
			       boardPtr->membase);
		}
	}

	return 0;

	printk("eggmcs913: unloading driver.\n");
fail_major:
	release_region(boardcfg.baseaddr, EGGMCS913_IOSIZE);
fail_region:
	if (boardcfg.fake != 0) {
		for (boardPtr = boardlist; boardPtr < boardlist + nboards;
		     boardPtr++) {
			if (boardPtr->membase != NULL) {
				free_pages(boardPtr->membase, boardPtr->sz);
			}
		}
	}

	kfree(boardlist);

	return stat;
}

void cleanup_module(void)
{
	eggmcs913_t* bPtr;

	if (boardcfg.irqhandler != NULL) {
		free_irq(boardcfg.irq, NULL);
	}
	release_region(boardcfg.baseaddr, EGGMCS913_IOSIZE);
	
	if (unregister_chrdev(major, eggmcs913_name) != 0) {
		printk("eggmcs913: cleanup_module failed\n");
	} else {
		MCS_PRINTK("eggmcs913: cleanup_module succeeded\n");
	}

	if (boardcfg.fake != 0) {
		for (bPtr = boardlist; bPtr < boardlist + nboards; bPtr++) {
			if (bPtr->membase != NULL) {
				free_pages(bPtr->membase, bPtr->sz);
			}
		}
	}

	kfree(boardlist);
}
#endif /* MODULE */
