/*
 * Device Driver for EG&G ACE-MCS 913
 *
 *
 */


/* Resources:

  - MCS-board uses the following ports :
           0x292, 0x294, 0x296-0x29a 
    (according to the docs)
  - seems like 0x29a is not used !?

  - the memory is mapped to 0xD0000 and is 
    (4096*4 bytes) = 0x4000 bytes large.
  
  - IRQ 3 or 5 at the end of a pass

  that's the same for all (maximum 8) boards. They are distinguished
  by addressing them through a port. (see below)
*/

#ifdef __KERNEL__

/* cf. Linux Journal 3/96 p.13 */
#ifdef DEBUG_EGGMCS913
#define MCS_PRINTK(format, args...)  printk (format, ## args)
#else
#define MCS_PRINTK(format, args...)
#endif


#define EGGMCS913_BOARDS_MAX 8

#ifndef EGGMCS913_MAJOR
#define EGGMCS913_MAJOR 63
#endif

#ifndef EGGMCS913_IRQ
#define EGGMCS913_IRQ        0x5
#endif

#ifndef EGGMCS913_BASEADDR
#define EGGMCS913_BASEADDR   0x292
#endif

#ifndef EGGMCS913_MEMBASE
#define EGGMCS913_MEMBASE    0xD0000
#endif


#define EGGMCS913_IOSIZE 0x9

/*
 the board effectively uses only 16kBytes
 (4096 channels * 4 bytes/channel)
 24 bits (3 bytes) of data , 1 byte used as ROI (region of interest)
 by the EG&G mcs-software.
*/
#define EGGMCS913_BYTESPERCHANNEL 4
#define EGGMCS913_CHANNELS        4096


/* The timer for the dwelltime is clocked with 2 MHz. */
#define EGGMCS913_MIN_DWELLTIME 2
/* EGGMCS913_MAX_DWELLTIME = 2147418112  XXX ? */
#define EGGMCS913_MAX_DWELLTIME ( ( 0x10000u/2 ) * 0x10000u)
#define EGGMCS913_MIN_PASSLENGTH 4
#define EGGMCS913_MAX_PASSLENGTH EGGMCS913_CHANNELS

/* This port is used to select 1 of max 8 MCS-boards installed.
  Board#1 (as selected by the jumpers) corresponds to 0 and so on. */
#define EGGMCS913_ADDRESS(base)    (base)

/* This port is used to start/stop the board and to give status
   information.

  The 8 bits are used as follows:

    write:                         read:

0  ext.start                     ext.stop.req
1  ext.start                     ext.start
2  ext.dwelltime                 ext.dwelltime
3  add-mode                      add-mode
4  scale0                        scale0
5  scale1                        scale1
6  scale2                        scale2
7  active (?)                    INTRQ

The scale[012] bits are used for the ramp option,
which was not available to me - no support.

*/
#define EGGMCS913_CONTROL(base)    (base + 0x02)

#define EGGMCS913_EXTSTOP      0x01
#define EGGMCS913_EXTSTART     0x03
#define EGGMCS913_EXTDWELLTIME 0x04
#define EGGMCS913_ADDMODE      0x08
/* At the moment I cannot exactly tell what it does, just use it
   the same way the MCS-software does. */
#define EGGMCS913_UNKNOWN_BIT  0x80
#define EGGMCS913_UNKNOWN_CMD  0x81


/*
   The 8253A timer-chip on the board is used as a counter to 
   generate channel-advances (dwelltime) and for the pass-length.
   Counter0 and Counter1 make up a 32 bit counter for the dwelltime,
   Counter2 is for the passlength and generates the midpass-signal.

   The addresses seem to be:
     0x296  counter2          ( ...0110 )
     0x297  control-word      ( ...0111 )
     0x298  counter0          ( ...1000 )
     0x299  counter1          ( ...1001 )

     only bit0 and bit1 are relevant for addressing the timer.

   (for detailed documentation get a book on PC-hardware)
*/
#define EGGMCS913_TIM8253_CW(base) (base + 0x05)    /*  0x297  */
#define EGGMCS913_TIM8253_C0(base) (base + 0x06)    /*  0x298  */
#define EGGMCS913_TIM8253_C1(base) (base + 0x07)    /*  0x299  */
#define EGGMCS913_TIM8253_C2(base) (base + 0x04)    /*  0x296  */

#define  LOWBYTE(x) ( x & 0xff)
#define HIGHBYTE(x) ( (x >> 8) & 0xff)

/* Data common to all boards */
typedef struct {
        int fake;
        int irq;
        unsigned long baseaddr;
        unsigned long* membase;

        int nchannels;
        int bytesperchannel;
        unsigned long memsize;

        void (*irqhandler)(int irq, void*, struct pt_regs *ignored);

} eggmcs913cfg_t;

/* Data specific to one board */
typedef struct {
	/* For fake mode each board needs its own memory. In real
	   operation this is identical for all boards and taken
	   from the config above. */
        unsigned long* membase;

        int boardid;
        int usecount;
        int controler;   /* There is a process, that has exclusive
			    control */

        long          dwelltime;
        int           cnt0;
        int           cnt1;
        int           passlength;
        unsigned long passcountpreset;
        int           extstart;
        int           extdwelltime;
        int           add;

        unsigned long passcount; 
        int active;
        int cmd;

        int stop_req;

	/* for fake mode */;
        int cur_channel;
	struct timer_list mcs_timer_list;
        int sz;
} eggmcs913_t;

/* Data per file-pointer */
typedef struct {
        eggmcs913_t *theBoard;
        int controler; 
        int startchannel;
        int size;
} EGGMCS913_ClientData;

#endif  /*  __KERNEL__ */

/* ioctl's */
#ifndef EGGMCS913_MAGIC
#define EGGMCS913_MAGIC    'E'
#endif

#define EGGMCS913_IOCTL_SET_PASSCOUNTPRESET  _IOW(EGGMCS913_MAGIC, 1, int)
#define EGGMCS913_IOCTL_SET_DWELLTIME        _IOW(EGGMCS913_MAGIC, 2, unsigned long)
#define EGGMCS913_IOCTL_SET_PASSLENGTH       _IOW(EGGMCS913_MAGIC, 3, int)
#define EGGMCS913_IOCTL_START                _IO(EGGMCS913_MAGIC, 4)
#define EGGMCS913_IOCTL_STOP                 _IO(EGGMCS913_MAGIC, 5)
#define EGGMCS913_IOCTL_STATUS               _IOR(EGGMCS913_MAGIC, 6, int)
#define EGGMCS913_IOCTL_CLEAR                _IO(EGGMCS913_MAGIC, 7)
#define EGGMCS913_IOCTL_GET_PASSCOUNT        _IOR(EGGMCS913_MAGIC, 8, unsigned long)
#define EGGMCS913_IOCTL_SET_EXTDWELLTIME     _IOW(EGGMCS913_MAGIC, 9, int)
#define EGGMCS913_IOCTL_SET_EXTSTART         _IOW(EGGMCS913_MAGIC, 10, int)
#define EGGMCS913_IOCTL_SET_ADD              _IOW(EGGMCS913_MAGIC, 11, int)
#define EGGMCS913_IOCTL_GET_EXTDWELLTIME     _IOR(EGGMCS913_MAGIC, 12, int)
#define EGGMCS913_IOCTL_GET_EXTSTART         _IOR(EGGMCS913_MAGIC, 13, int)
#define EGGMCS913_IOCTL_GET_ADD              _IOR(EGGMCS913_MAGIC, 14, int)
#define EGGMCS913_IOCTL_ACTIVE               _IOR(EGGMCS913_MAGIC, 15, int)
#define EGGMCS913_IOCTL_GET_DWELLTIME        _IOR(EGGMCS913_MAGIC, 16, long)
#define EGGMCS913_IOCTL_STOPREQ              _IO(EGGMCS913_MAGIC, 17)
#define EGGMCS913_IOCTL_NCHANNELS            _IOR(EGGMCS913_MAGIC, 18, int)
#define EGGMCS913_IOCTL_STARTCHANNEL         _IOW(EGGMCS913_MAGIC, 19, int)
#define EGGMCS913_IOCTL_MMAP                 _IO(EGGMCS913_MAGIC, 20)
#define EGGMCS913_IOCTL_MUNMAP               _IO(EGGMCS913_MAGIC, 21)
